/* bob.c - the main routine */
/*
	Copyright (c) 1991, by David Michael Betz
    BOB+ Copyright (c) 2005-2005 by R,-Erik Ebert
	All rights reserved
*/

#include <stdio.h>
#include <setjmp.h>
#include "bob.h"
#include <dos.h>
#include <dir.h>
#ifndef BP_RUN
  #define USE_COMPILE
#endif

static char* BANNER	=
#ifndef BP_RUN
      #ifdef BP_COMP
         "Bob+ v1.1d Compiler - Copyright (c) 2005-2010, by Ralf-Erik Ebert\n"
         "Parts are based on code of Bob, Copyright (c) 1991, by David M. Betz\n\n"
         "usage:\n"
         "bpc [-i][-d][-e [-I]] [sourcefile [..]] [-r objfile [..]] [-o outfile] \n"
         "options:\n"
         "    -d ... debug information during compilation\n"
         "    -i ... info (shows banner, syntax and names of compiled functions)\n"
         "    -e ... build standalone exe-file (link against bpr.exe runtime module)\n"
         "    -I ... link against bpri.exe rather than bpr.exe (only if -e set too)\n"
         "    -r ... specifies precompiled files to be loaded\n"
         "    -o ... name of output file\n";
      #else
         #ifdef USE_FLOAT
            "Bob+ v1.1d - Copyright (c) 2005-2010, by Ralf-Erik Ebert\n"
         #else
            "Bob+ v1.1di - Copyright (c) 2005-2010, by Ralf-Erik Ebert\n"
         #endif
         "Parts are based on code of Bob, Copyright (c) 1991, by David Betz\n\n"
         "usage:\n"
         #ifdef USE_FLOAT
            "bp [-i][-d][-t][-c] [-e [-I]] [sourcefile [..]] [-r objfile [..]] \n"
         #else
            "bpi [-i][-d][-t][-c] [-e] [sourcefile [..]] [-r objfile [..]] \n"
         #endif
         "   [-o outfile] [# userarg [..]]\n"
         "options:\n"
         "    -d ... debug information during compilation\n"
         "    -t ... trace bytecodes during execution\n"
         "    -i ... info (shows banner, syntax and names of compiled functions)\n"
         "    -c ... compile only\n"
 #ifdef USE_FLOAT        
         "    -e ... build standalone exe-file\n"
         "           (link against bpr.exe runtime module, only if -c set too)\n"
         "    -I ... link against bpri.exe rather than bpr.exe (only if -e set too)\n"
 #else
         "    -e ... build standalone exe-file\n"
         "           (link against bpri.exe runtime module, only if -c set too)\n"
#endif
         "    -r ... specifies precompiled files to be loaded\n"
         "    -o ... name of output file (only used with -c)\n"
         "     # ... marks begin of user argumentl ist\n";
      #endif  /* ! BP_COMP */
#else  /* BP_RUN */
   #ifdef USE_FLOAT
      "Bob+ runtime v1.1d - Copyright (c) 2005-2010, by Ralf-Erik Ebert\n"
   #else
      "Bob+ runtime v1.1di - Copyright (c) 2005-2010, by Ralf-Erik Ebert\n"
   #endif
      "Parts are based on code of Bob, Copyright (c) 1991, by David M. Betz\n\n"
      "usage:\n"
   #ifdef USE_FLOAT
      "bpr [-i][objfile [..]] [# userarg [..]]\n"
   #else
      "bpri [-i][objfile [..]] [# userarg [..]]\n"
   #endif
   "options:\n"
   "    -i ... info (shows banner and syntax)\n"
   "     # ... marks begin of user argumentlist\n";
#endif
/* global variables */

jmp_buf error_trap;

/* external variables */
extern int decode,trace;
extern int usrarg0;
extern int cmdargc;
extern char** cmdargv;
extern VALUE* sp;

int verbose = 0;
int compile = 0;
int buildexe = 0;
#ifdef USE_FLOAT
   int linkint = 0;
#else
   int linkint = 1;
#endif
int load = 0;
char* outname = "out.bpm";


#ifdef BP_RUN
   char _arg0[MAXPATH];
   
   char* getProgName()
   {
      int intvar;
      char far* penv;
      if ((_argv[0] != NULL) && (strlen(_argv[0]) > 0)) return _argv[0]; 
      intvar = *(int far*) MK_FP(_psp,0x2c); /* get environment segment */
      penv = (char far*) MK_FP(intvar,0); /* get environment pointer */
      while ((penv != NULL) && ((*penv) != '\0'))
      {
          while ((*penv) != '\0') ++penv;
          ++penv;
      }
      penv += 3;
      for (intvar = 0; intvar < MAXPATH; ++intvar, ++penv)
      {
         _arg0[intvar] = *penv;
         if ((*penv) == '\0') break;
      }
      return _arg0;
    }
#endif

/* main - the main routine */
main(argc,argv)
  int argc; char *argv[];
{
	char* bpstack;
    char fullname[20];
    int i,n;
    int firstsrc = 0, srccnt = 0;
#ifdef BP_RUN
    char* pname;
/*      char far* fptr1;
    char cmdbuf[128]; */
    FILE* fp = NULL;
    struct {
       long fpos;
       char tag[4];
    } tagstruct;
#endif
    /* initialize */
    bpstack = getenv("BPSTACK");
    i = 0;
    if (bpstack != 0)
    {
		i = atoi(bpstack);
	}
	if (i == 0) i = SMAX;
    initialize(i,CMAX);
    usrarg0 = argc;
    cmdargc = argc;
    cmdargv = argv;
#ifdef BP_RUN
/*    printf("arg0: %s\n",_argv[0]);
    fptr1 = MK_FP(_psp,0x80);
    n = (int) (char) *fptr1;
    for(i=0;i<n;++i) cmdbuf[i] = *(++fptr1);
    cmdbuf[i] = 0;
    printf("cmd: %s\n",cmdbuf); */
    pname = searchpath(getProgName());
    if (pname)
       fp = fopen(pname,"rb");
    if (fp)
    {
       fseek(fp,-8,SEEK_END);
       fread(&tagstruct,8,1,fp);
       tagstruct.tag[3] = 0;
       if (strcmp(tagstruct.tag,"BPR") == 0)
       {
          usrarg0 = 1;
          fseek(fp,tagstruct.fpos,SEEK_SET);
          if (!readModuleFP(fp))
          {
              printf("error reading module %s\n",argv[0]);
              return 1;
          }
          if (!execute("main"))
             return 1;
           i = 0;
           if (sp->v_type == DT_INTEGER)
              i = sp->v.v_integer;
           return i;
       }
       else
         fclose(fp);
    }
#endif
    /* load and execute some code */
    if (argc == 1) verbose = 1;
    for (i = 1; i < argc; ++i)
    if (strcmp(argv[i],"#") == 0)
    {
		usrarg0 = i+1;
		break;
	}
	else if (argv[i][0] == '-')
	{
		switch (argv[i][1])
		{
		case 'i':
		   verbose = 1;
		   break;
#ifndef BP_RUN
		case 'd':
		   decode = 1;
		   break;
  #ifndef BP_COMP
		case 't':
		   trace = 1;
		   break;
		case 'c':
		   compile = 1;
		   break;
  #endif	/* !BP_COMP */
		case 'e':
		   buildexe = 1;
		   break;
#ifdef USE_FLOAT		   
		case 'I':
		   linkint= 1;
		   break;
#endif		   
		case 'o':
		   if (i<argc-1)
		   {
			   i++;
			   outname = argv[i];
		   }
		   break;
		case 'r':
#else
        default:
#endif // !BP_RUN
		   while (i < argc-1)
		   {
			   if ( (argv[i+1][0] == '-') || (argv[i+1][0] == '#'))
			     break;
			   strcpy(fullname,argv[i+1]);
 		       if (strchr(fullname,'.') == 0)
				  strcat(fullname,".bpm");
			   if (!readModule(fullname))
			   {
				   printf("error reading module %s\n",fullname);
				   return 1;
			   }
			   i++;
		   }
#ifndef BP_RUN
         break;
		default:
		   printf("unknown option %s ignored\n",argv[i]);
#endif
		}
	}
	else
	{
		if (srccnt == 0) firstsrc = i;
		srccnt++;
	}
#ifdef BP_COMP
       if (srccnt > 0) compile=1;
#endif
	for (i=firstsrc;i < firstsrc+srccnt;i++)
	{
#ifndef BP_RUN
	    strcpy(fullname,argv[i]);
		if (strchr(fullname,'.') == 0)
		   strcat(fullname,".bp");
		if(!compile_file(fullname))
		   return 255;
#else
	   strcpy(fullname,argv[i]);
	   if (strchr(fullname,'.') == 0)
		  strcat(fullname,".bpm");
	   if (!readModule(fullname))
	   {
		   printf("error reading module %s\n",fullname);
		   return 1;
	   }
#endif
	}
	if (verbose)
	     osputs(BANNER);
#ifndef BP_RUN
	if (compile)
	{
  #ifndef BP_DEMO
           if (buildexe)
           {
              i = (writeModule("out.tmp")) ? 0 : 1;
              if (i == 0)
              {
                 if (strcmp(outname,"out.bpm") == 0)
                    outname = "out.exe";
                 i = (build_exe()) ? 0 : 1;  
              }
              
           }
           else
	      i = (writeModule(outname)) ? 0 : 1;
  #endif
    }
#endif 
#ifndef BP_COMP
	else
	{
       if (!execute("main"))
          return 1;
       i = 0;
       if (sp->v_type == DT_INTEGER)
          i = sp->v.v_integer;
    }
#endif
    return i;
}

#ifdef USE_COMPILE
/* compile_file - compile definitions in a file */
int compile_file(name)
  char *name;
{
    FILE *ifp;
    int result = FALSE;
    if ((ifp = fopen(name,"r")) != NULL) {
	result = compile_definitions(fgetc,ifp);
	fclose(ifp);
    }
    return result;
}

unsigned long cpy_file(src, dest, append)
char *src, *dest;
int append;
{
   FILE *fps, *fpd;
    char buf[256];
    int c;
   unsigned long flen = 0;
   fps = fopen(src,"rb");
   if (fps == NULL)
   {
       error("Unable to open file ('%s')",src);
       return 0;
   }
   fpd = fopen(dest,append ? "ab" : "wb");
   if (fpd == NULL)
   {
       fclose(fps);
       error("Unable to open file ('%s')",dest);
       return 0;
   }
   while(!feof(fps))
   {
      c = fgetc(fps);
      if (c == EOF) break;
      if (fputc(c,fpd) == EOF)
      {
         fclose(fps);
         fclose(fpd);
         error("Error while writing '%s'.",outname);
      }
      ++flen;
   }
   fclose(fps);
   fclose(fpd);
   return flen;
}
/* build exe from compiled file */
int build_exe()
{
    char* rtname, *rtpath;
    FILE* fp;
    unsigned len;
    unsigned long flen, dlen;
    int result = FALSE;
    rtname = linkint ? "bpri.exe" : "bpr.exe";
    rtpath = searchpath(rtname);
    if (rtpath == NULL)
          error("Unable to locate runtime file ('%s')",rtname);
    flen = cpy_file(rtpath,outname,FALSE);
    dlen = cpy_file("out.tmp",outname,TRUE);
    fp = fopen(outname,"r+b");
    if (fp == 0)
    {
       error("Open output file ('%s') failed.",outname);
    }
    fseek(fp,0,SEEK_END);
       fwrite(&flen,1,sizeof(unsigned long),fp);
       fwrite("BPR",1,4,fp);
/*       if (fseek(fp,2,SEEK_SET) == 0)
       {
          flen += dlen + 8;
          dlen = ((flen/512) << 16) + (flen % 512);
          if (dlen & 0xFFFF)
             dlen += 0x10000L;
          fwrite(&dlen,1,sizeof(unsigned long),fp);
       } */
       fclose(fp); 
    return TRUE;
}
#endif

/* info - display progress information */
info(fmt,a1,a2,a3,a4,a5,a6)
  char *fmt;
{
    if (!(decode || trace || verbose))  return;
	osputs("[ ");
    printf(fmt,a1,a2,a3,a4,a5,a6);
	osputs(" ]\n");
}

/* error - print an error message and exit */
error(fmt,a1,a2,a3,a4)
  char *fmt;
{
	osputs("Error: ");
	printf(fmt,a1,a2,a3,a4);
	osputs("\n");
    longjmp(error_trap,1);
}

osputc(ch)
  int ch;
{
    putc(ch,stdout);
}

osputs(str)
  char *str;
{
    fputs(str,stdout);
}
