/* bobdbg.c - debug routines */
/*
	Copyright (c) 1991, by David Michael Betz
	All rights reserved
*/
#ifndef BP_RUN
#include "bob.h"

/* instruction output formats */
#define FMT_NONE	0
#define FMT_BYTE	1
#define FMT_WORD	2
#define FMT_LIT		3

#ifndef POFOBUILD
typedef struct { char *ot_name; unsigned char ot_fmt; } OTDEF;
OTDEF otab[] = {
{  "(unknown)", FMT_NONE    },
{	"brt",		FMT_WORD	},
{	"brf",		FMT_WORD	},
{	"br",		FMT_WORD	},
{	"nil",		FMT_NONE	},
{	"push",		FMT_NONE	},
{	"not",		FMT_NONE	},
{	"neg",		FMT_NONE	},
{	"add",		FMT_NONE	},
{	"sub",		FMT_NONE	},
{	"mul",		FMT_NONE	},
{	"div",		FMT_NONE	},
{	"rem",		FMT_NONE	},
{	"band",		FMT_NONE	},
{	"bor",		FMT_NONE	},
{	"xor",		FMT_NONE	},
{	"bnot",		FMT_NONE	},
{	"shl",		FMT_NONE	},
{	"shr",		FMT_NONE	},
{	"lt",		FMT_NONE	},
{	"le",		FMT_NONE	},
{	"eq",		FMT_NONE	},
{	"ne",		FMT_NONE	},
{	"ge",		FMT_NONE	},
{	"gt",		FMT_NONE	},
{	"inc",		FMT_NONE	},
{	"dec",		FMT_NONE	},
{	"lit",		FMT_LIT		},
{	"return",	FMT_NONE	},
{	"call",		FMT_BYTE	},
{	"ref",		FMT_LIT		},
{	"set",		FMT_LIT		},
{	"vref",		FMT_NONE	},
{	"vset",		FMT_NONE	},
{	"mref",		FMT_BYTE	},
{	"mset",		FMT_BYTE	},
{	"aref",		FMT_BYTE	},
{	"aset",		FMT_BYTE	},
{	"tref",		FMT_BYTE	},
{	"tset",		FMT_BYTE	},
{	"tspace",	FMT_BYTE	},
{	"send",		FMT_BYTE	},
{	"dup2",		FMT_NONE	},
{	"new",		FMT_NONE	},
{	"delete",		FMT_NONE	}
};
#endif
/* decode_procedure - decode the instructions in a code object */
decode_procedure(code)
  VALUE *code;
{
#ifndef POFOBUILD
    int len,lc,n;
    len = code[1].v.v_string->s_length;
    for (lc = 0; lc < len; lc += n)
	n = decode_instruction(code,lc);
#endif
}

/* decode_instruction - decode a single bytecode instruction */
int decode_instruction(code,lc)
  VALUE *code; int lc;
{
#ifndef POFOBUILD
    char name[TKNSIZE+1]; //,buf[100];
    unsigned char *cp;
    OTDEF *op;
    int n=1;

    /* get bytecode pointer for this instruction */
    cp = code[1].v.v_string->s_data + lc;

    /* show the address and opcode */
    if (code[2].v_type == DT_CLASS) {
	printf("%s::",code[2].v.v_class->cl_name);
	//osputs(buf);
    }
    getcstring(name,sizeof(name),code[3].v.v_string);
    printf("%s %04x %02x ",name,lc,*cp);
    //osputs(buf);

    /* display the operands */
    if (*cp < sizeof(otab) / sizeof(OTDEF))
	{
		op = &otab[*cp];
	    switch (op->ot_fmt)
	    {
	    case FMT_NONE:
		printf("      %s\n",op->ot_name);
		// osputs(buf);
		break;
	    case FMT_BYTE:
		printf("%02x    %s %02x\n",cp[1],op->ot_name,cp[1]);
		//osputs(buf);
		n ++;
		break;
	    case FMT_WORD:
		printf("%02x %02x %s %02x%02x\n",cp[1],cp[2],
			op->ot_name,cp[2],cp[1]);
		//osputs(buf);
		n += 2;
		break;
	    case FMT_LIT:
		printf("%02x    %s %02x ; ",cp[1],op->ot_name,cp[1]);
		// osputs(buf);
		print1(TRUE,&code[cp[1]]);
		osputs("\n");
		n ++;
		break;
	    }
	    return (n);
	}

    /* unknown opcode */
    printf("      (UNKNOWN)\n");
    //osputs(buf);
#else
   return 0;
#endif
}
#endif