/* bobfcn.c - built-in classes and functions */
/*
	Copyright (c) 1991, by David Michael Betz
	All rights reserved
*/
#include <time.h>
#include <string.h>
#include <dos.h>
#include "bob.h"
#ifdef USE_FLOAT
  #include <math.h>
#endif

#define USE_GRAPHIC
#define USE_DOSFUNC
#ifndef BP_RUN
   #define USE_COMPILE
#endif

static union REGS _regs;
static struct SREGS _sregs;

int usrarg0=0;
int cmdargc=0;
char** cmdargv=0;
/* argument check */

#ifndef BP_COMP

int argcount(n,cnt)
int n,cnt;
{
	if (n < cnt)
	{
      error("not enough arguments");
      return FALSE;
    }
    if (n > cnt)
    {
		error("too many arguments");
		return FALSE;
	}
	return TRUE;
}

#endif

/* external variables */
extern DICTIONARY *symbols;
extern DICT_ENTRY* sym_end;

#ifndef BP_COMP

/* forward declarations */
int xnewvector(),xnewstring(),xprint();
int xfopen(),xfclose(),xgetc(),xputc();
int xstrlen(), xstrsize(), xvecsize(), xcompile(), xloadmodule();
int xstrcmp(), xstricmp();
int xgets(), xfgets(), xfputs(), xfeof(), xfreadval(), xfwriteval(), xstrsplit();
int xVec(), xint(), xstring(), xfree(), xmemsize();
int xgettypename(), xgettype(), xgetclassname(), xdynamic_cast();
int xrand(), xsrand(), xtime(), xtimer(), xctime(), xlocaltime();
int xgetargs(), xgetusrargs(), xargcnt(),xarg() ;
int xsetreg(), xgetreg(), xsegread(), xint86(), xint86x(), xaddr();
int xgetmemb(), xgetmemw(), xgetmemdw(), xsetmemb(), xsetmemw(), xsetmemdw(),xmovedata();
int xinportb(), xoutportb(), xinport(), xoutport();
#ifdef USE_FLOAT
int xsin(), xcos(), xtan(), xatan(), xsqrt(), xlog(), xexp(), xabs();
int xfloat(), xpow();
#endif
int xsetscrmode(), xsetpixel(), xline();
extern int loadlib();
extern int getproc();
extern int calldll();
extern int freelib();

int send();

#endif  /* ! BP_COMP */

typedef struct funcentry
{
	char* name;
	int (*fptr)();
} FUNCENTRY;

static FUNCENTRY functable[] =
{
#ifndef BP_COMP
	{"newvector",xnewvector},
	{"newstring",xnewstring},
    {"fopen",xfopen},
    {"fclose",xfclose},
    {"getc",xgetc},
    {"putc",xputc},
    {"print",xprint},
// extensions
    {"strlen",xstrlen},
    {"strsize",xstrsize},
    {"strcmp",xstrcmp},
    {"stricmp",xstricmp},
    {"vecsize",xvecsize},
#ifdef USE_COMPILE
    {"compile",xcompile},
#endif
#ifndef BP_DEMO
    {"loadmodule",xloadmodule},
#endif
    {"gets",xgets},
    {"fgets",xfgets},
    {"fputs",xfputs},
#ifndef BP_DEMO
    {"freadval",xfreadval},
    {"fwriteval",xfwriteval},
#endif
    {"feof",xfeof},
    {"strsplit",xstrsplit},
    {"Vec",xVec},
    {"T",xVec},
    {"int",xint},
    {"string",xstring},
    {"gettypename",xgettypename},
    {"gettype",xgettype},
    {"getclassname",xgetclassname},
    {"dynamic_cast",xdynamic_cast},
    {"free",xfree},
    {"memsize",xmemsize},
    {"rand",xrand},
    {"srand",xsrand},
    {"time",xtime},
    {"timer",xtimer},
    {"ctime",xctime},
    {"localtime",xlocaltime},
    {"getargs",xgetargs},
    {"getusrargs",xgetusrargs},
    {"argcnt",xargcnt},
    {"arg",xarg},
    {"addr",xaddr},
#ifdef USE_DOSFUNC
    {"setreg",xsetreg},
    {"getreg",xgetreg},
    {"segread",xsegread},
    {"int86",xint86},
    {"int86x",xint86x},
    {"getmemb",xgetmemb},
    {"getmemw",xgetmemw},
    {"getmemdw",xgetmemdw},
    {"setmemb",xsetmemb},
    {"setmemw",xsetmemw},
    {"setmemdw",xsetmemdw},
    {"movedata",xmovedata},
    {"inportb",xinportb},
    {"outportb",xoutportb},
    {"inport",xinport},
    {"outport",xoutport},
#endif
#ifdef USE_FLOAT
    {"float",xfloat},
    {"sin",xsin},
    {"cos",xcos},
    {"tan",xtan},
    {"atan",xatan},
    {"sqrt",xsqrt},
    {"log",xlog},
    {"exp",xexp},
    {"abs",xabs},
    {"pow",xpow},
#endif
#ifdef USE_GRAPHIC
    {"setscrmode",xsetscrmode},
    {"setpixel",xsetpixel},
    {"line",xline},
#endif
    {"LoadLibrary",loadlib},
    {"GetProcAddress",getproc},
    {"CallLibFunc",calldll},
    {"FreeLibrary",freelib}
#else /* BP_COMP - compiler needs not really implemented functions */

	{"newvector",NULL},
	{"newstring",NULL},
    {"fopen",NULL},
    {"fclose",NULL},
    {"getc",NULL},
    {"putc",NULL},
    {"print",NULL},
// extensions
    {"strlen",NULL},
    {"strsize",NULL},
    {"strcmp",NULL},
    {"stricmp",NULL},
    {"vecsize",NULL},
#ifdef USE_COMPILE
    {"compile",NULL},
#endif
#ifndef BP_DEMO
    {"loadmodule",NULL},
#endif
    {"gets",NULL},
    {"fgets",NULL},
    {"fputs",NULL},
#ifndef BP_DEMO
    {"freadval",NULL},
    {"fwriteval",NULL},
#endif
    {"feof",NULL},
    {"strsplit",NULL},
    {"Vec",NULL},
    {"T",NULL},
    {"int",NULL},
    {"string",NULL},
    {"gettypename",NULL},
    {"gettype",NULL},
    {"getclassname",NULL},
    {"dynamic_cast",NULL},
    {"free",NULL},
    {"memsize",NULL},
    {"rand",NULL},
    {"srand",NULL},
    {"time",NULL},
    {"timer",NULL},
    {"ctime",NULL},
    {"localtime",NULL},
    {"getargs",NULL},
    {"getusrargs",NULL},
    {"argcnt",NULL},
    {"arg",NULL},
    {"addr",NULL},
#ifdef USE_DOSFUNC
    {"setreg",NULL},
    {"getreg",NULL},
    {"segread",NULL},
    {"int86",NULL},
    {"int86x",NULL},
    {"getmemb",NULL},
    {"getmemw",NULL},
    {"getmemdw",NULL},
    {"setmemb",NULL},
    {"setmemw",NULL},
    {"setmemdw",NULL},
    {"movedata",NULL},
    {"inportb",NULL},
    {"outportb",NULL},
    {"inport",NULL},
    {"outport",NULL},
#endif
#ifdef USE_FLOAT
    {"float",NULL},
    {"sin",NULL},
    {"cos",NULL},
    {"tan",NULL},
    {"atan",NULL},
    {"sqrt",NULL},
    {"log",NULL},
    {"exp",NULL},
    {"abs",NULL},
    {"pow",NULL},
#endif
#ifdef USE_GRAPHIC
    {"setscrmode",NULL},
    {"setpixel",NULL},
    {"line",NULL},
#endif
    {"LoadLibrary",NULL},
    {"GetProcAddress",NULL},
    {"CallLibFunc",NULL},
    {"FreeLibrary",NULL}
#endif /* BP_COMP */
};

typedef struct typeentry
{
	int type;
	char* name;
} TYPEENTRY;

static TYPEENTRY typetable[] =
{
   { DT_NIL, "null" },
   { DT_CLASS, "class" },
   { DT_OBJECT, "object" },
   { DT_VECTOR, "vector" },
   { DT_INTEGER, "int" },
   { DT_STRING, "string" },
   { DT_BYTECODE, "bytecode" },
   { DT_CODE, "code" },
   { DT_VAR, "var" },
   { DT_FILE, "FILE" },
   { DT_FLOAT, "float" }
};
/* init_functions - initialize the internal functions */
void init_functions()
{
	int i,n;
	add_file("stdin",stdin);
    add_file("stdout",stdout);
    add_file("stderr",stderr);
n = sizeof(functable) / sizeof(FUNCENTRY);
	for (i=0;i<n;i++) add_function(functable[i].name,functable[i].fptr);
	sym_end = symbols->di_contents;
}

/* add_function - add a built-in function */
static add_function(name,fcn)
  char *name; int (*fcn)();
{
    DICT_ENTRY *sym;
    sym = addentry(symbols,name,ST_SFUNCTION);
    set_code(&sym->de_value,fcn);
}

/* add_file - add a built-in file */
static add_file(name,fp)
  char *name; FILE *fp;
{
    DICT_ENTRY *sym;
    sym = addentry(symbols,name,ST_SDATA);
    set_file(&sym->de_value,fp);
}

#ifndef BP_COMP /* skip all functions for compile only version */

static int xmemsize(argc)
   int argc;
{
	unsigned m =0;
	if (argc == 1)
	{
		chktype(0,DT_INTEGER);
		m = sp->v.v_integer;
		++sp;
	}
	else
	{
	   argcount(argc,0);
   }
	if (m != 0L)
	{
		asm {
			push ax
			push bx
	        pushf
			mov ah, 48H
			mov bx, 0xffff
			int 21H
			mov m, bx
			popf
			pop bx
			pop ax
		}
		 set_integer(&sp[0],(long)m * 16L);

	}
	else
	   set_integer(&sp[0],(unsigned)coreleft());
}
static int xfree(argc)
  int argc;
{
/*	STRING* dtorstr; */
    argcount(argc,1);
    if ((sp->v_type == DT_STRING) ||
        (sp->v_type == DT_VECTOR) ||
        (sp->v_type == DT_OBJECT)
       )
       {
          free(sp->v.v_object);
	   }
    ++sp;
    set_nil(&sp[0]);
}

static int xint(argc)
  int argc;
{
    long iarg; /* char buf[11]; */
    argcount(argc,1);
    if (sp->v_type == DT_INTEGER)
    {
       iarg = sp->v.v_integer;
    }
#ifdef USE_FLOAT
    else if (sp->v_type == DT_FLOAT)
    {
       iarg = sp->v.v_float;
    }
#endif
    else if (sp->v_type == DT_STRING)
    {
	/*	getcstring(buf,11,sp->v.v_string);
		iarg = strtol(buf,0,0); */
		iarg = strtol(sp->v.v_string->s_data,0,0);
	}
	else
	   error("invalid type: %d",sp->v_type);
    ++sp;
    set_integer(&sp[0],iarg);
}

/* string(val[,formatstring]) */
static int xstring(argc)
  int argc;
{
    char *buf = NULL; /* fmt[100]; */
    char* fmstr = NULL;
    void far* data = NULL;
    int bytes = 30;
    FILE* fp;
    if (argc != 2)
      argcount(argc,1);
    else
    {
		chktype(0,DT_STRING);
		/* getcstring(fmt,100,sp->v.v_string); */
		fmstr = sp->v.v_string->s_data;
		bytes += strlen(fmstr);
		++sp;
	}
    if (sp->v_type == DT_INTEGER)
    {
		if (!fmstr) fmstr = "%ld";
		data = sp->v.v_far;
    }
#ifdef USE_FLOAT
    else if (sp->v_type == DT_FLOAT)
    {
		if (!fmstr) fmstr = "%f";
		data = sp->v.v_far;
    }
#endif
    else if (sp->v_type == DT_STRING)
    {
		if (!fmstr) fmstr = "%s";
		data = MK_FP(_DS,sp->v.v_string->s_data);
		bytes += strlen(sp->v.v_string->s_data);
	}
	else
	   error("invalid type: %d",sp->v_type);
    ++sp;
/*	if (_osmajor < 3)
	  bytes = 1023;
	else
	{
	   fp = fopen("NUL","wb");
	   bytes = fprintf(fp,fmstr,data);
	   fclose(fp);
	   if (bytes == EOF)
	       error("string: error calculating needed buffer");
	}
*/	
	buf = getmemory(bytes+1);
	sprintf(buf,fmstr,data);
    set_string(&sp[0],makestring(buf));
    free(buf);
}

#ifdef USE_FLOAT

static int xfloat(argc)
  int argc;
{
    float farg; /* char buf[20]; */
    argcount(argc,1);
    if (sp->v_type == DT_INTEGER)
    {
       farg = sp->v.v_integer;
    }
    else if (sp->v_type == DT_FLOAT)
    {
       farg = sp->v.v_float;
    }
    else if (sp->v_type == DT_STRING)
    {
		/* getcstring(buf,20,sp->v.v_string); */
		sscanf(sp->v.v_string->s_data,"%f",&farg);
	}
	else
	   error("invalid type: %d",sp->v_type);
    ++sp;
    set_float(&sp[0],farg);
}

typedef double (*MFType) (double);
static int mathfunc(argc,f)
  int argc; MFType f;
{
    float arg;
    argcount(argc,1);
    if (sp->v_type == DT_INTEGER)
       arg = sp->v.v_integer;
    else
    {
       chktype(0,DT_FLOAT);
       arg = sp->v.v_float;
    }
    ++sp;
    set_float(&sp[0],f(arg));
}

static int xsin(argc)
  int argc;
{
	mathfunc(argc,sin);
}

static int xcos(argc)
  int argc;
{
	mathfunc(argc,cos);
}

static int xtan(argc)
  int argc;
{
	mathfunc(argc,tan);
}

static int xatan(argc)
  int argc;
{
	mathfunc(argc,atan);
}

static int xsqrt(argc)
  int argc;
{
	mathfunc(argc,sqrt);
}

static int xlog(argc)
  int argc;
{
	mathfunc(argc,log);
}

static int xexp(argc)
  int argc;
{
	mathfunc(argc,exp);
}

static int xpow(argc)
  int argc;
{
    float arg1, arg2;
    argcount(argc,2);
    if (sp->v_type == DT_INTEGER)
       arg2 = sp->v.v_integer;
    else
    {
       chktype(0,DT_FLOAT);
       arg2 = sp->v.v_float;
    }
    if (sp[1].v_type == DT_INTEGER)
       arg1 = sp[1].v.v_integer;
    else
    {
       chktype(1,DT_FLOAT);
       arg1 = sp[1].v.v_float;
    }
    sp += 2;
    set_float(&sp[0], exp(arg2 * log(arg1)));
}

static int xabs(argc)
  int argc;
{
    float arg; long iarg;
    argcount(argc,1);
    if (sp->v_type == DT_INTEGER)
    {
       iarg = sp->v.v_integer;
       ++sp;
       set_integer(&sp[0],labs(iarg));
    }
   else
    {
       chktype(0,DT_FLOAT);
       arg = sp->v.v_float;
	   ++sp;
	   set_float(&sp[0],(float)fabs(arg));
    }
}

#endif // USE_FLOAT
/* xnewvector - allocate a new vector */
static int xnewvector(argc)
  int argc;
{
    unsigned size;
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    size = sp->v.v_integer;
    ++sp;
    set_vector(&sp[0],newvector(size));
}

/* xVec - allocate a new initialized */
static int xVec(argc)
  int argc;
{
	VALUE* vec; int i;
    int size = argc;
    vec = newvector(size);
    for(i=0; i<size; i++)
       vec[i+1] = sp[size-i-1];
    sp += argc;
    set_vector(&sp[0],vec);
}

#ifdef USE_COMPILE
/* xcompile - load and compile sourcecode at runtime */
static int xcompile(argc)
  int argc;
{
	/* char name[255]; */
    int result;
    argcount(argc,1);
    chktype(0,DT_STRING);
    /* getcstring(name,sizeof(name),sp[0].v.v_string); */
    result = compile_file(sp->v.v_string->s_data);
    ++sp;
    set_integer(&sp[0],result);
}

#endif

#ifndef BP_DEMO
/* xloadmodule - load a precompiled module at runtime */
static int xloadmodule(argc)
  int argc;
{
	/* char name[255]; */
	char* name  = NULL;
	char* fname;
    int result;
    argcount(argc,1);
    chktype(0,DT_STRING);
    name = sp[0].v.v_string->s_data;
  /*  getcstring(name,sizeof(name),sp[0].v.v_string); */
	fname = searchpath(name);
	if (fname == 0)
		fname = name;
	result = readModule(fname);
    ++sp;
    set_integer(&sp[0],result);
}
#endif

/* xstrlen - calculate string length */
static int xstrlen(argc)
  int argc;
{
    int size;
    argcount(argc,1);
    chktype(0,DT_STRING);
    size = strlen(sp->v.v_string->s_data);
    ++sp;
    set_integer(&sp[0],size);
}

/* xstrlen - calculate string size */
static int xstrsize(argc)
  int argc;
{
    int size;
    argcount(argc,1);
    chktype(0,DT_STRING);
    size = sp->v.v_string->s_length;
    ++sp;
    set_integer(&sp[0],size);
}



/* xstrcmp - compare strings */
static int xxstrcmp(argc,cmpfunc)
  int argc,(*cmpfunc)(const char*,const char*);
{
   /* int tmp; */
    char* b1; char*b2;
    argcount(argc,2);
    chktype(0,DT_STRING);
    chktype(1,DT_STRING);
    b2 = sp[0].v.v_string->s_data;
    b1 = sp[1].v.v_string->s_data;
/*    tmp = sp->v.v_string->s_length + 1;
    b2 = (char*)malloc(tmp);
    getcstring(b2,tmp,sp[0].v.v_string);
    tmp = sp[1].v.v_string->s_length + 1;
    b1 = (char*)malloc(tmp);
    getcstring(b1,tmp,sp[1].v.v_string); */
    sp+=2;
    set_integer(&sp[0],cmpfunc(b1,b2));
 /*   free(b1);
    free(b2); */
}

/* xstrcmp - compare strings */
static int xstrcmp(argc)
  int argc;
{
	xxstrcmp(argc,strcmp);
}

/* xstrcmp - compare strings */
static int xstricmp(argc)
  int argc;
{
	xxstrcmp(argc,stricmp);
}

/* xstrsplit - make tokenvector from string */
static int xstrsplit(argc)
  int argc;
{
    int size, i; char* buf; char *delim; char* p;
    VALUE* vec;
    argcount(argc,2);
    chktype(0,DT_STRING); // delimiter
    chktype(1,DT_STRING); // string
    delim = sp[0].v.v_string->s_data;
  /*  getcstring(delim,sizeof(delim),sp[0].v.v_string); */
    size = sp[1].v.v_string->s_length+1;
    buf = getmemory(size);
    getcstring(buf,size,sp[1].v.v_string);
    // count tokens
    for(i=0, p=strtok(buf,delim);p != 0;i++,p=strtok(0,delim));
    vec = newvector(i);
    getcstring(buf,size,sp[1].v.v_string);
    for(i=0, p=strtok(buf,delim); p!=0; i++,p=strtok(0,delim))
    {
		set_string(&vec[i+1],makestring(p));
	}
	free(buf);
    sp += 2;;
    set_vector(&sp[0],vec);
}

static int xgetargs(argc)
   int argc;
{
	int i;
	VALUE* vec;
	argcount(argc,0);
	vec = newvector(cmdargc);
	for(i=0;i<cmdargc;i++)
	{
		set_string(&vec[i+1],makestring(cmdargv[i]));
	}
    set_vector(&sp[0],vec);
}

static int xgetusrargs(argc)
   int argc;
{
	int i,n;
	VALUE* vec;
	argcount(argc,0);
	n = cmdargc - usrarg0;
	vec = newvector(n);
	for(i=usrarg0;i<cmdargc;i++)
	{
		set_string(&vec[i-usrarg0+1],makestring(cmdargv[i]));
	}
    set_vector(&sp[0],vec);
}

static int xargcnt(argc)
   int argc;
{
    set_integer(&sp[0],fp[2].v.v_integer);
}

static int xarg(argc)
   int argc;
{
    int idx;
    chktype(0,DT_INTEGER);
    idx = sp->v.v_integer;
    if ((idx < 0) || (idx >= fp[2].v.v_integer))
       error("arg(%d) : index out of bounds",idx);
    ++sp;
    *sp = fp[fp[2].v.v_integer-idx+2];
}

/* xvecsize - calculate size of vector */
static int xvecsize(argc)
  int argc;
{
    unsigned size;
    argcount(argc,1);
    chktype(0,DT_VECTOR);
    size = sp->v.v_vector->v.v_integer;
    ++sp;
    set_integer(&sp[0],size);
}

/* xnewstring - allocate a new string */
static int xnewstring(argc)
  int argc;
{
    int size;
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    size = sp->v.v_integer;
    ++sp;
    set_string(&sp[0],newstring(size));
}

/* xfopen - open a file */
static int xfopen(argc)
  int argc;
{
/*    char name[200],mode[10]; */
    FILE *fp;
    argcount(argc,2);
    chktype(0,DT_STRING);
    chktype(1,DT_STRING);
/*    getcstring(name,sizeof(name),sp[1].v.v_string);
    getcstring(mode,sizeof(mode),sp[0].v.v_string); */
    fp = fopen(sp[1].v.v_string->s_data,sp[0].v.v_string->s_data);
    sp += 2;
    if (fp)
	set_file(&sp[0],fp);
    else
	set_nil(&sp[0]);
}

/* xfclose - close a file */
static int xfclose(argc)
  int argc;
{
    int sts;
    argcount(argc,1);
    chktype(0,DT_FILE);
    sts = fclose(sp[0].v.v_fp);
    ++sp;
    set_integer(&sp[0],sts);
}

/* xfeof - check if end of filr reached */
static int xfeof(argc)
  int argc;
{
    int sts;
    argcount(argc,1);
    chktype(0,DT_FILE);
    sts = feof(sp[0].v.v_fp);
    ++sp;
    set_integer(&sp[0],sts);
}

/* xgetc - get a character from a file */
static int xgetc(argc)
  int argc;
{
    int ch;
    argcount(argc,1);
    chktype(0,DT_FILE);
    ch = getc(sp[0].v.v_fp);
    ++sp;
    set_integer(&sp[0],ch);
}

/* xgets - get a string from stdin */
static int xgets(argc)
  int argc;
{
    char buffer[200];
    argcount(argc,0);
    if (!gets(buffer))
       buffer[0] = '\0';
    set_string(&sp[0],makestring(buffer));
}

/* xfgets - get a string from file */
static int xfgets(argc)
  int argc;
{
    char buffer[200];
    argcount(argc,1);
    chktype(0,DT_FILE);
    if (!fgets(buffer,200,sp[0].v.v_fp))
       buffer[0] = '\0';
    ++sp;
    set_string(&sp[0],makestring(buffer));
}

/* xputc - output a character to a file */
static int xputc(argc)
  int argc;
{
    int ch;
    argcount(argc,2);
    chktype(0,DT_FILE);
    chktype(1,DT_INTEGER);
    ch = putc(sp[1].v.v_integer,sp[0].v.v_fp);
    sp += 2;
    set_integer(&sp[0],ch);
}

/* xputc - output a string to a file */
static int xfputs(argc)
  int argc;
{
	int ch; /* size, */
	/* nchar* buf; */
    argcount(argc,2);
    chktype(0,DT_FILE);
    chktype(1,DT_STRING);
/*    size = sp[1].v.v_string->s_length+1;
    buf = (char*) malloc(size);
    getcstring(buf,size,sp[1].v.v_string); */
    ch = fputs(sp[1].v.v_string->s_data,sp[0].v.v_fp);
/*    free(buf); */
    sp += 2;
    set_integer(&sp[0],ch);
}
#ifndef BP_DEMO
/* xfwriteval - write a value to a binary file */
static int xfwriteval(argc)
  int argc;
{
	int result;
    argcount(argc,2);
    chktype(0,DT_FILE);
    result = writeSingleValue(sp[0].v.v_fp,&sp[1]);
    sp += 2;
    set_integer(&sp[0],result);
}

/* xfwriteval - read a value from a binary file */
static int xfreadval(argc)
  int argc;
{
	FILE* fp;
    argcount(argc,1);
    chktype(0,DT_FILE);
    fp = sp[0].v.v_fp;
    sp++;
    set_nil(sp);
    if (!readSingleValue(fp,sp))
       set_nil(sp);
}
#endif // !BP_DEMO
/* xtime - get time since 00:00:00 GMT 1.1.1970 in seconds */
static int xtime(argc)
  int argc;
{
    argcount(argc,0);
    set_integer(&sp[0],time(0));
}

/* xtimer - get time since program invocation in milliseconds */
static int xtimer(argc)
  int argc;
{
	clock_t c;
	long ms;
    argcount(argc,0);
    c=clock();
#ifdef USE_FLOAT
    ms = c * 1000;
    ms /= CLK_TCK;
#else
    ms = c * 10000;
    ms /= 182;
#endif
    set_integer(&sp[0],ms);
}

/* xctime - convert time value to string */
static int xctime(argc)
  int argc;
{
	char* buf;
	time_t t;
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    t = sp[0].v.v_integer;
    buf=ctime(&t);
    sp++;
    set_string(&sp[0],makestring(buf));
}

/* xctime - get localized time as vector of integers:
    sec, min, hour, mday, mon, year, wday, yday, isdst
   */
static int xlocaltime(argc)
  int argc;
{
	time_t t;
	struct tm *lt;
	int i, *p;
	VALUE* vec;
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    t = sp[0].v.v_integer;
    lt=localtime(&t);
    vec=newvector(9);
    p = (int*) lt;
    for(i=1;i<=9;i++,p++)
        set_integer(&vec[i],*p);
    sp++;
    set_vector(&sp[0],vec);
}

/* xrand - get random number */
static int xrand(argc)
  int argc;
{
    argcount(argc,0);
    set_integer(&sp[0],rand());
}

/* xsrand - reinitialize random number generator*/
static int xsrand(argc)
  int argc;
{
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    srand(sp[0].v.v_integer);
    set_integer(&sp[0],rand());
    sp++;
    set_nil(sp);
}

static int xgettypename(argc)
  int argc;
{
	int type;
    argcount(argc,1);
    chktype(0,DT_INTEGER);
    type = sp[0].v.v_integer;
    if ((type < _DTMIN) || (type > _DTMAX))
    	error("undefined type: %d",type);
    sp++;
    set_string(&sp[0],makestring(typetable[type].name));
}

static int xgettype(argc)
  int argc;
{
	int type;
    argcount(argc,1);
    type = sp[0].v_type;
    sp++;
    set_integer(sp,type);
}

static int xgetclassname(argc)
   int argc;
{
	VALUE* val;
	CLASS* cl;
    argcount(argc,1);
    val = &sp[0];
    cl = (val->v_type == DT_CLASS) ? val->v.v_class : ((val->v_type == DT_OBJECT) ? val->v.v_object[0].v.v_class : 0);
	sp++;
	if (cl == 0)
	  set_nil(sp);
	else
	  set_string(sp,makestring(cl->cl_name));
}

static int xdynamic_cast(argc)
   int argc;
{
	VALUE val;
	CLASS* cl;
    argcount(argc,2);
    chktype(1,DT_CLASS);
    val = sp[0];
   // cl = (val.v_type == DT_CLASS) ? val.v.v_class : ((val.v_type == DT_OBJECT) ? val.v.v_object[0].v.v_class : 0);
    cl = (val.v_type == DT_OBJECT) ? val.v.v_object[0].v.v_class : 0;
    while (cl != 0)
    {
		if (cl == sp[1].v.v_class)
		   break;
		cl = cl->cl_base;
	}
	if (cl == 0)
	  set_nil(&val);
	sp++;
	sp[0] = val;
}

/* xprint - generic print function */
static int xprint(argc)
  int argc;
{
    int n;
    for (n = argc; --n >= 0; )
	print1(FALSE,&sp[n]);
    sp += argc;
    set_nil(sp);
}

#endif /* ! BP_COMP */
static char* printFormats[] = {
	"", "(Class %s)", "(Object %lx)", "(Vector %lx)", "%ld",
	"%s", "(Bytecode %lx)", "(Code %lx)", "%s::%s", "(Vector %lx)", "%g"
};
/* print1 - print one value */
print1(qflag,val)
  int qflag; VALUE *val;
{
	int vtype;char c;
    char *p;
    CLASS *bpclass;
    int len;
    vtype = val->v_type;
    if ((vtype < _DTMIN) || (vtype > _DTMAX))
      	error("undefined type: %d",val->v_type);
    p = printFormats[vtype];
    switch (vtype)
    {
    case DT_NIL:
	  printf("null");
	break;
    case DT_CLASS:
	printf(p,val->v.v_class->cl_name);
	break;
#ifdef USE_FLOAT
    case DT_FLOAT:
    printf(p,(double)(val->v.v_float));

    break;
#endif
    case DT_STRING:
	if (qflag) osputc('"');
	osputs(val->v.v_string->s_data);
	if (qflag) osputc('"');
	break;
    case DT_VAR:
	if ((bpclass = val->v.v_var->de_dictionary->di_class) == NULL)
	    printf(p,"",val->v.v_var->de_key);
	else
	    printf(p,bpclass->cl_name,val->v.v_var->de_key);
	break;
    default:
       printf(p,val->v);
    }
}

#ifndef BP_COMP

/* ============================================================ */
/*                  BIOS/DOS functions                          */
/* ============================================================ */


#define SWITCHREGSET(NAME,PREFIX) if (stricmp(#NAME,buf) == 0) PREFIX##.##NAME = sp->v.v_integer; else
static int xsetreg(argc)
   int argc;
{
	char buf[6];
	argcount(argc,2);
	chktype(1,DT_STRING);
	chktype(0,DT_INTEGER);
	getcstring(buf,6,sp[1].v.v_string);
	SWITCHREGSET(ax,_regs.x)
	SWITCHREGSET(bx,_regs.x)
	SWITCHREGSET(cx,_regs.x)
	SWITCHREGSET(dx,_regs.x)
	SWITCHREGSET(si,_regs.x)
	SWITCHREGSET(di,_regs.x)
	SWITCHREGSET(cflag,_regs.x)
	SWITCHREGSET(flags,_regs.x)
	SWITCHREGSET(al,_regs.h)
	SWITCHREGSET(ah,_regs.h)
	SWITCHREGSET(bl,_regs.h)
	SWITCHREGSET(bh,_regs.h)
	SWITCHREGSET(cl,_regs.h)
	SWITCHREGSET(ch,_regs.h)
	SWITCHREGSET(dl,_regs.h)
	SWITCHREGSET(dh,_regs.h)
	SWITCHREGSET(ds,_sregs)
	SWITCHREGSET(es,_sregs)
	SWITCHREGSET(ss,_sregs)
	SWITCHREGSET(cs,_sregs)
	{
	   error("undefined register name: %s",buf);
	}
	sp+=2;
	set_nil(&sp[0]);
}

static int xaddr(argc)
   int argc;
{
	void* result;
	argcount(argc,1);
	switch (sp->v_type)
	{
	case DT_STRING:
	   result=sp->v.v_string->s_data;
	   break;
	case DT_INTEGER:
	case DT_FLOAT:
	   result = &(sp->v.v_integer);
	   break;
	default:
	  result = sp->v.v_near;
	}
    sp++;
	 set_farptr(&sp[0],MK_FP(_DS,(unsigned)result));
}

#ifdef USE_DOSFUNC

#define SWITCHREGGET(NAME,PREFIX) if (stricmp(#NAME,buf) == 0) result = PREFIX##.##NAME ; else
static int xgetreg(argc)
   int argc;
{
	char buf[6];
	unsigned result;
	argcount(argc,1);
	chktype(0,DT_STRING);
	getcstring(buf,6,sp[0].v.v_string);
	SWITCHREGGET(ax,_regs.x)
	SWITCHREGGET(bx,_regs.x)
	SWITCHREGGET(cx,_regs.x)
	SWITCHREGGET(dx,_regs.x)
	SWITCHREGGET(si,_regs.x)
	SWITCHREGGET(di,_regs.x)
	SWITCHREGGET(cflag,_regs.x)
	SWITCHREGGET(flags,_regs.x)
	SWITCHREGGET(al,_regs.h)
	SWITCHREGGET(ah,_regs.h)
	SWITCHREGGET(bl,_regs.h)
	SWITCHREGGET(bh,_regs.h)
	SWITCHREGGET(cl,_regs.h)
	SWITCHREGGET(ch,_regs.h)
	SWITCHREGGET(dl,_regs.h)
	SWITCHREGGET(dh,_regs.h)
	SWITCHREGGET(ds,_sregs)
	SWITCHREGGET(es,_sregs)
	SWITCHREGGET(ss,_sregs)
	SWITCHREGGET(cs,_sregs)
	{
	   error("undefined register name: %s",buf);
	}
	sp++;
    set_integer(&sp[0],result);
}


static int xsegread(argc)
   int argc;
{
	argcount(argc,0);
	segread(&_sregs);
	set_nil(&sp[0]);
}

static int xint86(argc)
   int argc;
{
	int n;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	n = sp->v.v_integer;
	sp++;
    set_integer(&sp[0],int86(n,&_regs,&_regs));
}

static int xint86x(argc)
   int argc;
{
	int n;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	n = sp->v.v_integer;
	sp++;
    set_integer(&sp[0],int86x(n,&_regs,&_regs,&_sregs));
}

static int xgetmemb(argc)
   int argc;
{
	 unsigned char far* p;
	unsigned char result;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	p = (unsigned char far*) (sp->v.v_integer);
	result = *p;
	sp++;
        set_integer(&sp[0],result);
}

static int xgetmemw(argc)
   int argc;
{
	unsigned int far* p;
	unsigned int result;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	p = (unsigned int far*) (sp->v.v_integer);
	result = *p;
	sp++;
        set_integer(&sp[0],result);
}

static int xgetmemdw(argc)
   int argc;
{
	 unsigned long far* p;
	unsigned long result;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	p = (unsigned long far*) (sp->v.v_integer);
	result = *p;
	sp++;
        set_integer(&sp[0],result);
}

static int xsetmemb(argc)
   int argc;
{
	argcount(argc,2);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	*((unsigned char far*)(sp[1].v.v_integer)) = (unsigned char)(sp[0].v.v_integer);
	sp+=2;
    set_nil(&sp[0]);
}

static int xsetmemw(argc)
   int argc;
{
	argcount(argc,2);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	*((unsigned int far*)(sp[1].v.v_integer)) = (unsigned int)(sp[0].v.v_integer);
	sp+=2;
    set_nil(&sp[0]);
}

static int xsetmemdw(argc)
   int argc;
{
	argcount(argc,2);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	*((unsigned long far*)(sp[1].v.v_integer)) = (unsigned long)(sp[0].v.v_integer);
	sp+=2;
    set_nil(&sp[0]);
}

static int xmovedata(argc)
   int argc;
{
        unsigned seg_s, ofs_s, seg_d, ofs_d;
        size_t cnt;
	argcount(argc,3);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	chktype(2,DT_INTEGER);
	cnt = sp[0].v.v_integer;
	seg_d = FP_SEG(sp[1].v.v_far);
	ofs_d = FP_OFF(sp[1].v.v_far);
	seg_s = FP_SEG(sp[2].v.v_far);
	ofs_s = FP_OFF(sp[2].v.v_far);
	movedata(seg_s,ofs_s,seg_d,ofs_d,cnt);
	sp+=3;
    set_nil(&sp[0]);
}

static int xinportb(argc)
   int argc;
{
	int port;
	unsigned char result;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	port = sp->v.v_integer;
	result = inportb(port);
	sp++;
    set_integer(&sp[0],result);
}


static int xinport(argc)
   int argc;
{
	int port;
	int result;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	port = sp->v.v_integer;
	result = inport(port);
	sp++;
    set_integer(&sp[0],result);
}

static int xoutportb(argc)
   int argc;
{
	argcount(argc,2);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	outportb(sp[1].v.v_integer,sp[0].v.v_integer);
	sp+=2;
    set_nil(&sp[0]);
}

static int xoutport(argc)
   int argc;
{
	argcount(argc,2);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	outport(sp[1].v.v_integer,sp[0].v.v_integer);
	sp+=2;
    set_nil(&sp[0]);
}

#endif /* USE_DOSFUNC */

#ifdef USE_GRAPHIC
static int xsetscrmode(argc)
   int argc;
{
	union REGS regs;
	argcount(argc,1);
	chktype(0,DT_INTEGER);
	regs.h.ah=0;
	regs.h.al=sp->v.v_integer;
	int86(0x10,&regs,&regs);
	++sp;
	set_nil(&sp[0]);
}


static int xsetpixel(argc)
   int argc;
{
	union REGS regs;
	argcount(argc,3);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	chktype(2,DT_INTEGER);
	regs.h.ah=0x0c;
	regs.h.al=sp[0].v.v_integer;
	regs.x.dx=sp[1].v.v_integer;
	regs.x.cx=sp[2].v.v_integer;
	int86(0x10,&regs,&regs);
	sp+=3;
	set_nil(&sp[0]);
}

static int xline(argc)
   int argc;
{
	union REGS regs, r2;
	int x1,y1,x2,y2,dx,dy,dist,incx,incy,xerr,yerr;
	register int i;
	argcount(argc,5);
	chktype(0,DT_INTEGER);
	chktype(1,DT_INTEGER);
	chktype(2,DT_INTEGER);
	chktype(3,DT_INTEGER);
	chktype(4,DT_INTEGER);
	regs.h.ah=0x0c;
	regs.h.al=sp[0].v.v_integer;
	x1 = sp[4].v.v_integer;
	y1 = sp[3].v.v_integer;
	x2 = sp[2].v.v_integer;
	y2 = sp[1].v.v_integer;

	dx = x2-x1;
	dy = y2-y1;

	/* increment and sign for both directions */
	if (dx < 0)
	{
		incx=-1;
		dx=-dx;
	}
	else
	   incx = (dx != 0) ? 1 : 0;

	if (dy < 0)
	{
		incy=-1;
		dy=-dy;
	}
	else
	   incy = (dy != 0) ? 1 : 0;

	/* largest distance */
	dist = (dx>dy) ? dx : dy;

	/* init */
	regs.x.cx = x1;
	regs.x.dx = y1;
	xerr = dx;
	yerr = dy;

	/* draw pixels */
	for (i=0; i<dist; i++)
	{
		int86(0x10,&regs,&r2);

		xerr += dx;
		yerr += dy;

		if (xerr>dist)
		{
			xerr -= dist;
			regs.x.cx += incx;
		}
		if (yerr>dist)
		{
			yerr -= dist;
			regs.x.dx += incy;
		}
	}

	regs.x.cx=x2;
	regs.x.dx=y2;
	int86(0x10,&regs,&r2);
	sp+=5;
	set_nil(&sp[0]);
}

#endif /* USE_GRAPHIC */

#endif /* ! BP_COMP */
